/*
** Module   : ConversionProfile.java
** Abstract : Enumeration of conversion configuration profiles for use with runtime conversion.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20140405 Created initial version for use with ConversionPool.
*/

package com.goldencode.p2j.persist;

/**
 * An enumeration of conversion configuration profiles which can be loaded and performed by the
 * pattern engine at runtime to convert dynamic temp-tables and queries.
 */
enum ConversionProfile
{
   /** Convert schema AST to p2o AST */
   P2O("schema/p2o", false),
   
   /** Generate Hibernate ORM document */
   HIBERNATE("schema/hibernate", false),
   
   /** Create DMO interface and implementation class */
   JAVA_DMO("schema/java_dmo", false),
   
   /** Annotations */
   ANNOTATIONS("annotations/annotations", true),
   
   /** Business logic base structure */
   BASE_STRUCTURE("convert/base_structure", true),
   
   /** Core conversion */
   CORE_CONVERSION("convert/core_conversion", true),
   
   /** Anti-parse Java AST to source code */
   BREW("convert/brew", false),
   
   /** Fix-up an in-memory FIND query */
   FIX_FIND("convert/fix_in_mem_find_query", false),
   
   /** Fix-up an in-memory query */
   FIX_QUERY("convert/fix_in_mem_query", false);
   
   /** Profile name */
   private final String name;
   
   /** Flag indicating whether conditional rule-set processing should be enabled */
   private final boolean conditional;
   
   /**
    * Constructor.
    * 
    * @param   name
    *          Profile name.
    * @param   conditional
    *          Flag indicating whether conditional rule-set processing should be enabled.
    */
   private ConversionProfile(String name, boolean conditional)
   {
      this.name = name;
      this.conditional = conditional;
   }
   
   /**
    * Get the string representation of this enum value.
    * 
    * @return  String representation.
    */
   public String toString()
   {
      return name;
   }
   
   /**
    * Get the profile name.
    * 
    * @return  Profile name.
    */
   String getName()
   {
      return name;
   }
   
   /**
    * Get the conditional flag.
    * 
    * @return  Conditional flag.
    */
   boolean isConditional()
   {
      return conditional;
   }
}
